import json

from crownstone_core.packets.behaviour.ActiveDays import ActiveDays
from crownstone_core.packets.behaviour.BehaviourTypes import BehaviourType, DAY_START_TIME_SECONDS_SINCE_MIDNIGHT
from crownstone_core.packets.behaviour.PresenceDescription import BehaviourPresence
from crownstone_core.packets.behaviour.TimeDescription import BehaviourTimeContainer, BehaviourTime, BehaviourTimeType
from crownstone_core.util.DataStepper import DataStepper
from crownstone_core.util.fletcher import fletcher32_uint8Arr


def DEFAULT_ACTIVE_DAYS():
    return ActiveDays()


def DEFAULT_TIME():
    return BehaviourTimeContainer(
        BehaviourTime().fromType(BehaviourTimeType.afterSunset),
        BehaviourTime().fromType(BehaviourTimeType.afterSunrise),
    )



class BehaviourBase:

    def __init__(self, profileIndex=None, behaviourType=BehaviourType.behaviour, intensity=None, activeDays=None, time=None, presence=None, endCondition=None, idOnCrownstone=None):
        self.profileIndex = 0 if profileIndex is None else profileIndex
        self.behaviourType = behaviourType
        self.intensity = 100 if intensity is None else max(0, min(100, intensity))
        self.activeDays = DEFAULT_ACTIVE_DAYS() if activeDays is None else activeDays
        self.fromTime = DEFAULT_TIME().fromTime if time is None else time.fromTime
        self.untilTime = DEFAULT_TIME().untilTime if time is None else time.untilTime
        self.presence = presence
        self.endCondition = endCondition
        self.idOnCrownstone = idOnCrownstone

        self.valid = True


    def setDimPercentage(self, value):
        self.intensity = value
        return self

    def setTimeAllday(self, dayStartTimeSecondsSinceMidnight=DAY_START_TIME_SECONDS_SINCE_MIDNIGHT):
        self.fromTime  = BehaviourTime().fromType(BehaviourTimeType.afterMidnight,  dayStartTimeSecondsSinceMidnight)
        self.untilTime = BehaviourTime().fromType(BehaviourTimeType.afterMidnight, dayStartTimeSecondsSinceMidnight)
        return self
    
    def setTimeWhenDark(self):
        self.fromTime  = BehaviourTime().fromType(BehaviourTimeType.afterSunset)
        self.untilTime = BehaviourTime().fromType(BehaviourTimeType.afterSunrise)
        return self
    
    def setTimeWhenSunUp(self):
        self.fromTime  = BehaviourTime().fromType(BehaviourTimeType.afterSunrise)
        self.untilTime = BehaviourTime().fromType(BehaviourTimeType.afterSunset)
        return self
    
    def setTimeFromSunrise(self, offsetMinutes = 0):
        self.fromTime = BehaviourTime().fromType(BehaviourTimeType.afterSunrise, offsetSeconds=60*offsetMinutes)
        return self
    
    def setTimeFromSunset(self, offsetMinutes = 0):
        self.fromTime  = BehaviourTime().fromType(BehaviourTimeType.afterSunset, offsetSeconds=60*offsetMinutes)
        return self
    
    def setTimeToSunrise(self, offsetMinutes = 0):
        self.untilTime = BehaviourTime().fromType(BehaviourTimeType.afterSunrise, offsetSeconds=60*offsetMinutes)
        return self
    
    
    def setTimeToSunset(self, offsetMinutes = 0):
        self.untilTime = BehaviourTime().fromType(BehaviourTimeType.afterSunset, offsetSeconds=60 * offsetMinutes)
        return self
    
    def setTimeFrom(self, hours, minutes): 
        self.fromTime  = BehaviourTime().fromTime(hours, minutes)
        return self
    
    
    def setTimeTo(self, hours, minutes):
        self.untilTime  = BehaviourTime().fromTime(hours, minutes)
        return self
        

    """
    The payload is made up from
     - BehaviourType  1B
     - Intensity      1B
     - profileIndex   1B
     - ActiveDays     1B
     - From           5B
     - Until          5B

     - Presence       13B --> for Switch Behaviour and Smart Timer
     - End Condition  17B --> for Smart Timer
     """

    def fromData(self, data):
        payload = DataStepper(data)

        firstByte = payload.getUInt8()
        if not BehaviourType.has_value(firstByte):
            self.valid = False
            return self

        self.behaviourType = BehaviourType(firstByte)
        self.intensity = payload.getUInt8()
        self.profileIndex = payload.getUInt8()
        self.activeDays = ActiveDays().fromData(payload.getUInt8())
        self.fromTime = BehaviourTime().fromData(payload.getAmountOfBytes(5))  # 4 5 6 7 8
        self.untilTime = BehaviourTime().fromData(payload.getAmountOfBytes(5))  # 9 10 11 12 13

        if self.fromTime.valid == False or self.untilTime.valid == False:
            self.valid = False
            return self

        if self.behaviourType == BehaviourType.behaviour:
            if payload.length >= 14 + 13:
                self.presence = BehaviourPresence().fromData(
                    payload.getAmountOfBytes(13))  # 14 15 16 17 18 19 20 21 22 23 24 25 26
                if not self.presence.valid:
                    self.valid = False
                    return self
            else:
                self.valid = False
                return self

        if self.behaviourType == BehaviourType.smartTimer:
            if payload.length >= 14 + 13 + 17:
                presence = BehaviourPresence().fromData(payload.getAmountOfBytes(17))
                if not presence.valid:
                    self.valid = False
                    return self

                self.endCondition = presence

            else:
                self.valid = False
                return self

    def getPacket(self):
        arr = []

        arr.append(self.behaviourType.value)
        arr.append(self.intensity)
        arr.append(self.profileIndex)

        arr.append(self.activeDays.getMask())

        arr += self.fromTime.getPacket()
        arr += self.untilTime.getPacket()

        return arr

    def getHash(self):
        return fletcher32_uint8Arr(self._getPaddedPacket())

    def getDictionary(self, dayStartTimeSecondsSinceMidnight=DAY_START_TIME_SECONDS_SINCE_MIDNIGHT):
        typeString = "BEHAVIOUR"
        if self.behaviourType == BehaviourType.twilight:
            typeString = "TWILIGHT"

        dataDictionary = {}
        if self.behaviourType == BehaviourType.twilight:
            dataDictionary["action"] = {"type": "DIM_WHEN_TURNED_ON", "data": self.intensity}
            dataDictionary["time"] = self._getTimeDictionary(dayStartTimeSecondsSinceMidnight)

        else:
            # behaviour and smart timer have the same format
            dataDictionary["action"] = {"type": "BE_ON", "data": self.intensity}
            dataDictionary["time"] = self._getTimeDictionary(dayStartTimeSecondsSinceMidnight)

            if self.presence is not None:
                dataDictionary["presence"] = self.presence.getDictionary()

            if self.endCondition is not None:
                endConditionDictionary = {}
                endConditionDictionary["type"] = "PRESENCE_AFTER"
                endConditionDictionary["presence"] = self.endCondition.getDictionary()

                dataDictionary["endCondition"] = endConditionDictionary

        returnDict = {"type": typeString, "data": dataDictionary, "activeDays": self.activeDays.getDictionary(),
                      "idOnCrownstone": self.idOnCrownstone, "profileIndex": self.profileIndex}

        return returnDict

    def _getTimeDictionary(self, dayStartTimeSecondsSinceMidnight=DAY_START_TIME_SECONDS_SINCE_MIDNIGHT):
        returnDict = {}

        # check if always
        if self.fromTime.timeType == BehaviourTimeType.afterMidnight and self.fromTime.offset == dayStartTimeSecondsSinceMidnight and self.untilTime.timeType == BehaviourTimeType.afterMidnight and self.untilTime.offset == dayStartTimeSecondsSinceMidnight:
            returnDict["type"] = "ALL_DAY"
            return returnDict

        # its not always! construct the from and to parts.
        returnDict["type"] = "RANGE"
        returnDict["from"] = self.fromTime.getDictionary()
        returnDict["to"] = self.untilTime.getDictionary()

        return returnDict

    def _getPaddedPacket(self):
        packet = self.getPacket()
        if len(packet) % 2 != 0:
            packet.append(0)

        return packet


    def __str__(self):
        return json.dumps(self.getDictionary())


__author__ = 'Bart van Vliet'

import binascii
import struct

class Conversion:

	####################################
	# Conversions between uint and int #
	####################################
	@staticmethod
	def uint8_to_int8(byte):
		"""	Convert a uint8 to an int8 """
		res = byte
		if res > 127:
			res -= 256
		return res

	@staticmethod
	def int8_to_uint8(byte):
		"""	Convert an int8 to a uint8 """
		res = byte
		if res < 0:
			res += 256
		return res

	@staticmethod
	def uint16_to_int16(val):
		""" Convert a uint16 to an int16 """
		res = val
		if res > 32767:
			res -= 65536
		return res

	@staticmethod
	def int16_to_uint16(val):
		""" Convert an int16 to a uint16 """
		res = val
		if res < 0:
			res += 65536
		return res

	@staticmethod
	def uint32_to_int32(val):
		""" Convert an uint32 to a int32 """
		res = val
		if res > 2147483647:
			res -= 4294967296
		return res

	@staticmethod
	def int32_to_uint32(val):
		""" Convert an int32 to a uint32 """
		res = val
		if res < 0:
			res += 4294967296
		return res

	#######################################
	# Conversions between uint8 and float #
	#######################################
	@staticmethod
	def uint8_array_to_float(arr8):
		""" Convert an array of 4 bytes to a float """
		return struct.unpack('<f', bytearray(arr8))[0]

	@staticmethod
	def float_to_uint8_array(val):
		""" Convert a float to an array of 4 bytes """
		return bytearray(struct.pack('<f', val))

	########################################
	# Conversions between uint8 and uint16 #
	########################################
	@staticmethod
	def uint8_to_uint16(byte0, byte1):
		""" Convert 2 bytes to a uint16 """
		return (byte1 << 8) + byte0

	@staticmethod
	def uint8_array_to_uint16(arr8):
		""" Convert an array of 2 bytes to a uint16 """
		return (arr8[1] << 8) + arr8[0]

	@staticmethod
	def uint8_array_to_int16(arr8):
		""" Convert an array of 2 bytes to a uint16 """
		return Conversion.uint16_to_int16(Conversion.uint8_array_to_uint16(arr8))

	@staticmethod
	def uint16_to_uint8_array(value):
		""" Convert a number into an array of 2 bytes. """
		return [
			(value >> 0 & 0xFF),
			(value >> 8 & 0xFF)
		]

	@staticmethod
	def uint8_to_uint8_array(value):
		""" Convert a number into an array of 1 bytes. """
		return [ value & 0xFF ]

	@staticmethod
	def uint8_array_to_uint16_array(arr8):
		"""	Convert a uint8 array to a uint16 array	"""
		arr16 = []
		for i in range(1, len(arr8), 2):
#			arr16.append(Conversion.uint8_array_to_uint16(arr8[i-1 : i+1]))
			arr16.append(Conversion.uint8_to_uint16(arr8[i-1], arr8[i]))
		return arr16

	@staticmethod
	def uint16_array_to_uint8_array(arr16):
		"""	Convert a uint16 array to a byte array """
		arr8 = []
		for val in arr16:
			arr8.extend(Conversion.uint16_to_uint8_array(val))
		return arr8


	########################################
	# Conversions between uint8 and uint32 #
	########################################
	@staticmethod
	def uint8_to_uint32(byte0, byte1, byte2, byte3):
		""" Convert 4 bytes to a uint32 """
		return (byte3 << 24) + (byte2 << 16) + (byte1 << 8) + byte0

	@staticmethod
	def uint8_array_to_uint32(arr8):
		""" Convert an array of 4 bytes to a uint32 """
		return (arr8[3] << 24) + (arr8[2] << 16) + (arr8[1] << 8) + arr8[0]

	@staticmethod
	def uint8_array_to_uint64(arr8):
		""" Convert an array of 4 bytes to a uint32 """
		return (arr8[7] << 56) + \
			   (arr8[6] << 48) + \
			   (arr8[5] << 40) + \
			   (arr8[4] << 32) + \
			   (arr8[3] << 24) + \
			   (arr8[2] << 16) + \
			   (arr8[1] << 8) + \
			   arr8[0]


	@staticmethod
	def uint8_array_to_int32(arr8):
		""" Convert an array of 4 bytes to a uint32 """
		return Conversion.uint32_to_int32(Conversion.uint8_array_to_uint32(arr8))

	@staticmethod
	def uint8_array_to_uint32_array(arr8):
		"""	Convert a uint8 array to a uint16 array	"""
		arr32 = []
		for i in range(3, len(arr8), 4):
#			arr32.append(Conversion.uint8_array_to_uint32(arr8[i-3 : i+1]))
			arr32.append(Conversion.uint8_to_uint32(arr8[i-3], arr8[i-2], arr8[i-1], arr8[i]))
		return arr32

	@staticmethod
	def uint32_to_uint8_array(value):
		""" Convert a number into an array of 4 bytes. """
		return [
			(value >> 0 & 0xFF),
			(value >> 8 & 0xFF),
			(value >> 16 & 0xFF),
			(value >> 24 & 0xFF)
		]
	@staticmethod
	def int32_to_uint8_array(value):
		""" Convert a number into an array of 4 bytes. """
		return Conversion.uint32_to_uint8_array(Conversion.int32_to_uint32(value))

	@staticmethod
	def uint64_to_uint8_array(value):
		""" Convert a number into an array of 8 bytes. (little endian) """
		return [ value >> 8*i & 0xFF for i in range(8)]

	@staticmethod
	def uint32_array_to_uint8_array(arr32):
		"""	Convert a uint32 array to a byte array """
		arr8 = []
		for val in arr32:
			arr8.extend(Conversion.uint32_to_uint8_array(val))
		return arr8


	@staticmethod
	def uint32_to_uint16_reversed_array(uint32):
		return [
			uint32 >> 16,
			(uint32 >> 0) % (0xFFFF + 1),
		]

	@staticmethod
	def uint16_reversed_array_to_uint32(uint16Array):
		return uint16Array[1] + (uint16Array[0] * (0xFFFF + 1))


	##############################
	# uint8 array to/from string #
	##############################
	@staticmethod
	def uint8_array_to_string(arr8):
#		string = ""
#		for i in range(0, len(arr8)):
#			string += chr(arr8[i])
#		return string
#		return str(bytearray(arr8))
		return bytearray(arr8).decode()

	@staticmethod
	def string_to_uint8_array(string):
		arr8 = []
		for i in range(0, len(string)):
			arr8.append(ord(string[i]))
		return bytes(arr8)


	########################
	# Number to hex string #
	########################
	@staticmethod
	def uint8_to_hex_string(val):
		if val > 255 or val < 0:
			raise Exception("Value must be of type uint8")
		hex_str = "%02x" % val
		return hex_str

	@staticmethod
	def uint8_array_to_hex_string(arr):
#		hex_str = Conversion.uint8_to_hex_string(arr[0])
#		for i in range(1, len(arr)):
#			hex_str += Conversion.uint8_to_hex_string(arr[i])
#		return hex_str
		return binascii.b2a_hex(bytearray(arr)).decode('utf-8')

	@staticmethod
	def uint16_to_hex_string(val):
		arr8 = Conversion.uint16_to_uint8_array(val)
		return Conversion.uint8_array_to_hex_string(arr8)

	@staticmethod
	def uint32_to_hex_string(val):
		arr8 = Conversion.uint32_to_uint8_array(val)
		return Conversion.uint8_array_to_hex_string(arr8)


	#######################
	# Hex string to array #
	#######################
	
	@staticmethod
	def ibeaconUUIDString_to_uint8_array(inputStr):
		hexStr = inputStr.replace(":","").replace("-","")
		return Conversion.hex_string_to_uint8_array(hexStr)
	
	@staticmethod
	def ibeaconUUIDString_to_reversed_uint8_array(inputStr):
		arr = Conversion.ibeaconUUIDString_to_uint8_array(inputStr)
		arr.reverse()
		return arr
	
	@staticmethod
	def ascii_or_hex_string_to_16_byte_array(inputStr):
		if len(inputStr) == 16:
			return Conversion.string_to_uint8_array(inputStr)
		else :
			return Conversion.hex_string_to_uint8_array(inputStr)
	
	
	@staticmethod
	def hex_string_to_uint8_array(hexStr):
#		""" Convert a string which represents a hex byte buffer to an uint8 array """
#		""" Only converts buffer from start to start+size bytes """
#		buf = buffer_str.split(" ")
#		if (size == False):
#			size=len(buf)
#		arr = []
#		for i in range(start, start+size):
#			arr.append(int(buf[i], 16))
#		return arr
		"""
		Converts a hexadecimal string to a byte array
		:param hexStr: hexadecimal string to be converted
		:type hexStr: str
		:rtype: bytearray
		"""
		return bytearray(binascii.a2b_hex(hexStr.encode('utf-8')))

	@staticmethod
	def hex_string_to_uint16_array(hexStr):
#	def hex_string_to_uint16_array(buffer_str, start=0, size=False):
#		""" Convert a string which represents a hex byte buffer to an uint16 array """
#		""" Only converts buffer from start to start+size*2 bytes """
#		buf = buffer_str.split(" ")
#		if (size == False):
#			size=len(buf)/2
#		arr16 = []
#		for i in range(0, size):
#			arr16.append(Conversion.uint8_array_to_uint16([int(buf[start+2*i], 16), int(buf[start+2*i+1], 16)]))
#		return arr16
		"""
		Converts a hexadecimal string to a uint16 array
		:param hexStr: hexadecimal string to be converted
		:type hexStr: str
		:rtype: list
		"""
		arr8 = bytearray(binascii.a2b_hex(hexStr.encode('utf-8')))
		return Conversion.uint8_array_to_uint16(arr8)

	@staticmethod
	def hex_string_to_uint32_array(hexStr):
#	def hex_string_to_uint32_array(buffer_str, start=0, size=False):
#		""" Convert a string which represents a hex byte buffer to an uint32 array """
#		""" Only converts buffer from start to start+size*4 bytes """
#		buf = buffer_str.split(" ")
#		if (size == False):
#			size=len(buf)/4
#		arr32 = []
#		for i in range(0, size):
#			arr32.append(Conversion.uint8_array_to_uint32([int(buf[start+4*i], 16), int(buf[start+4*i+1], 16), int(buf[start+4*i+2], 16), int(buf[start+4*i+3], 16)]))
#		return arr32
		"""
		Converts a hexadecimal string to a uint32 array
		:param hexStr: hexadecimal string to be converted
		:type hexStr: str
		:rtype: list
		"""
		arr8 = bytearray(binascii.a2b_hex(hexStr.encode('utf-8')))
		return Conversion.uint8_array_to_uint32(arr8)

	@staticmethod
	def address_to_uint8_array(addressStr):
		"""
		Converts a bluetooth address string to a uint8 array
		:param addressStr: address string to be converted, must be like: "01:23:45:67:89:AB"
		:type addressStr: str
		:rtype: list
		"""
		hexStrArr = addressStr.split(":")
		if len(hexStrArr) != 6:
			return []

		arr8 = []
		for p in reversed(hexStrArr):
			arr8.append(Conversion.hex_string_to_uint8_array(p)[0])
		return bytearray(arr8)

	@staticmethod
	def uint8_array_to_address(arr8):
		"""
        Converts a uint8 array to a bluetooth address string
        :param arr8: array of uint8, should be 6 long.
        :type arr8: bytearray
        :rtype: str
        """
		if (len(arr8) != 6):
			return ""
		hexStr = Conversion.uint8_to_hex_string(arr8[-1])
		for b in reversed(arr8[0:-1]):
			hexStr = hexStr + ":" + Conversion.uint8_to_hex_string(b)
		return hexStr


	@staticmethod
	def bit_array_to_010_array(bitArray):
		result = [0] * len(bitArray)

		for i in range(0,len(bitArray)):
			if bitArray[i]:
				result[i] = 1

		return result



	@staticmethod
	def uint8_to_bit_array(val):
		result = [False, False, False, False, False, False, False, False]
		
		one = 1
		
		result[0] = (val & (one << 0)) != 0
		result[1] = (val & (one << 1)) != 0
		result[2] = (val & (one << 2)) != 0
		result[3] = (val & (one << 3)) != 0
		result[4] = (val & (one << 4)) != 0
		result[5] = (val & (one << 5)) != 0
		result[6] = (val & (one << 6)) != 0
		result[7] = (val & (one << 7)) != 0
		
		return result

	@staticmethod
	def uint16_to_bit_array(val):
		result = [False] * 16
		one = 1

		for i in range(0, 16):
			result[i] = (val & (one << 15 - i)) != 0

		return result

	@staticmethod
	def uint32_to_bit_array(val):
		result = [False] * 32
		one = 1
		
		for i in range(0,32):
			result[i] = (val & (one << 31 - i)) != 0
		
		return result
	
	@staticmethod
	def uint32_to_bit_array_reversed(val):
		result = [False] * 32
		one = 1
		
		for i in range(0,32):
			result[i] = (val & (one << 31 - i)) != 0
		
		return list(reversed(result))
	


from crownstone_core.util.fletcher import fletcher32_uint8Arr
import math

from crownstone_core.protocol.BluenetTypes import MicroappOpcode

# Object to be filled by the user. It contains a subset of the data required in the 
# packet, which e.g. also contains opcodes, checksum, etc.
class MicroappUploadCmd(object):
    def __init__(self, protocol, app_id, buf, chunk_size):
        self.protocol = protocol
        self.app_id = app_id
        self.buffer = buf
        self.size = len(buf)
        self.chunk_size = chunk_size

# Object to be filld by the user.
class MicroappRequestCmd(object):
    def __init__(self, protocol, app_id, buf, chunk_size):
        self.protocol = protocol
        self.app_id = app_id
        self.size = len(buf)
        self.count = math.ceil(self.size / chunk_size)
        self.chunk_size = chunk_size

# Object to be filld by the user. We need the entire buffer to calculate a checksum
class MicroappValidateCmd(object):
    def __init__(self, protocol, app_id, buf, chunk_size):
        self.protocol = protocol
        self.app_id = app_id
        self.buffer = buf 
        self.size = len(buf)
        self.chunk_size = chunk_size

# Object, convenient to be filled by the user.
class MicroappEnableCmd(object):
    def __init__(self, protocol, app_id, enable, offset):
        self.protocol = protocol
        self.app_id = app_id
        self.enable = enable
        self.param0 = offset



# All the packet fields required by the receiving Crownstone. 
class MicroappUploadPacket(object):
    def __init__(self, protocol, app_id, index, checksum, buf):
        self.protocol = protocol
        self.app_id = app_id
        self.opcode = MicroappOpcode.UPLOAD
        self.index = index
        self.checksum = checksum
        self.buffer = buf

# All the packet fields required by the receiving Crownstone. Expects a MicroappRequestCmd as input.
class MicroappRequestPacket(object):
    def __init__(self, command):
        self.protocol = command.protocol
        self.app_id = command.app_id
        self.opcode = MicroappOpcode.REQUEST
        self.size = command.size
        self.count = command.count
        self.chunk_size = command.chunk_size

# All the packet fields required by the receiving Crownstone. Expects a MicroappUploadCmd as input.
class MicroappValidatePacket(object):
    def __init__(self, command):
        self.protocol = command.protocol
        self.app_id = command.app_id
        self.opcode = MicroappOpcode.VALIDATE
        self.size = command.size
        self.checksum = 0xFF
        self.buffer = command.buffer

    def calculateChecksum(self):
        n = len(self.buffer)
        buf = bytearray(n)
        buf[0:n] = self.buffer[0:n]
        if len(buf) % 2 != 0:
            print("To check: do we make this even as well? For checksum calc")
            buf.append(0)
        self.checksum = fletcher32_uint8Arr(buf)
        print("LOG: checksum used: ", hex(self.checksum & 0xFFFF))

# All the packet fields required by the receiving Crownstone. Expects a MicroappEnableCmd as input.
class MicroappEnablePacket(object):
    def __init__(self, command):
        self.protocol = command.protocol
        self.app_id = command.app_id
        if command.enable:
            self.opcode = MicroappOpcode.ENABLE
        else:
            self.opcode = MicroappOpcode.DISABLE
        self.param0 = command.param0



class MicroappPacketInternal(object):

    def __init__(self, data):
        self.index = 0
        self.checksum = 0xCAFE
        self.count = math.ceil(data.size / data.chunk_size)
        self.chunk = bytearray(data.chunk_size)
        self.data = data
        if data.buffer:
            self.chunk[0 : data.chunk_size] = data.buffer[0 : data.chunk_size]

    def update(self):
        # if next is available, go to next index
        if (not self.nextAvailable()):
            return
        self.index += 1
        offset = self.index * self.data.chunk_size
        if (not self.last()):
            self.chunk[0 : self.data.chunk_size] = self.data.buffer[offset : offset + self.data.chunk_size]
        else:
            print("LOG: last piece")
            # Divide into remaining parts [data 0xFF]
            remaining0 = self.data.size - offset
            remaining1 = self.data.chunk_size - remaining0
            fill_buffer = bytearray([0xFF] * remaining1)
            self.chunk[0 : remaining0] = self.data.buffer[offset : offset + remaining0]
            self.chunk[remaining0 : self.data.chunk_size] = fill_buffer[0 : remaining1]
        print("LOG: size chunk is ", len(self.chunk))

    def last(self):
        offset = self.index * self.data.chunk_size
        if self.data.size > (offset + self.data.chunk_size):
            return False
        return True

    def nextAvailable(self):
        if ((self.index + 1) < self.count):
            return True
        return False

    def getPacket(self):
        self.calculateChecksum()
        packet = MicroappUploadPacket(self.data.protocol, self.data.app_id, self.index,
                self.checksum, self.chunk)
        return packet

    def calculateChecksum(self):
        # Needs padding if the default chunk can be odd-sized
        buf = bytearray(len(self.chunk))
        buf[0:len(self.chunk)] = self.chunk[0:len(self.chunk)]
        if len(buf) % 2 != 0:
            buf.append(0)
        self.checksum = fletcher32_uint8Arr(buf)
        print("LOG: checksum used: ", hex(self.checksum & 0xFFFF))



from crownstone_core.packets.serviceDataParsers.opCode3.opCode3_type0 import parseOpCode3_type0
from crownstone_core.packets.serviceDataParsers.opCode3.opCode3_type1 import parseOpCode3_type1
from crownstone_core.packets.serviceDataParsers.opCode3.opCode3_type2 import parseOpCode3_type2
from crownstone_core.packets.serviceDataParsers.opCode3.opCode3_type3 import parseOpCode3_type3
from crownstone_core.packets.serviceDataParsers.opCode4.opCode4_type0 import parseOpCode4_type0
from crownstone_core.packets.serviceDataParsers.opCode7.opCode7_type4 import parseOpCode7_type4
from crownstone_core.util.Conversion import Conversion


def parseOpCode3(serviceData, data):
    if len(data) == 16:
        serviceData.dataType = data[0]
        if serviceData.dataType == 0:
            parseOpCode3_type0(serviceData, data)
        elif serviceData.dataType == 1:
            parseOpCode3_type1(serviceData, data)
        elif serviceData.dataType == 2:
            parseOpCode3_type2(serviceData, data)
        elif serviceData.dataType == 3:
            parseOpCode3_type3(serviceData, data)
        else:
            parseOpCode3_type0(serviceData, data)






def parseOpCode4(serviceData, data):
    if len(data) == 16:
        serviceData.dataType = data[0]
        serviceData.setupMode = True
        if serviceData.dataType == 0:
            parseOpCode4_type0(serviceData, data)
        else:
            parseOpCode4_type0(serviceData, data)


def parseOpCode5(serviceData, data):
    if len(data) == 16:
        serviceData.dataType = data[0]

        if serviceData.dataType == 0:
            parseOpCode3_type0(serviceData, data)
        elif serviceData.dataType == 1:
            parseOpCode3_type1(serviceData, data)
        elif serviceData.dataType == 2:
            parseOpCode3_type2(serviceData, data)
            serviceData.rssiOfExternalCrownstone = Conversion.uint8_to_int8(data[15])
        elif serviceData.dataType == 3:
            parseOpCode3_type3(serviceData, data)
            serviceData.rssiOfExternalCrownstone = Conversion.uint8_to_int8(data[15])
        elif serviceData.dataType == 4:
            parseOpCode7_type4(serviceData, data)
        else:
            parseOpCode3_type0(serviceData, data)


def parseOpCode6(serviceData, data):
    if len(data) == 16:
        serviceData.dataType = data[0]
    
        serviceData.setupMode = True
        
        if serviceData.dataType == 0:
            parseOpCode4_type0(serviceData, data)
        else:
            parseOpCode4_type0(serviceData, data)
        





from crownstone_core.Exceptions import CrownstoneError
from crownstone_core.util.Conversion import Conversion

class BufferFiller :
    """
    Class to fill a data buffer.

    Data will be appended to the buffer for each consecutive set call.

    Function calls will raise an exception on error.
    """
    def __init__(self, data=None):
        """
        Constructor

        data : initial list of bytes, can be None.
        """
        if data is None:
            data = []
        self.data = data

    def getLength(self):
        """ Get the current length of the buffer in bytes. """
        return len(self.data)

    def getBuffer(self):
        """ Get the buffer. """
        return self.data

    def putUInt8(self, value):
        """ Append a uint8 to the buffer. """
        self.data.append(value)

    def putInt8(self, value):
        """ Append an int8 to the buffer. """
        self.putUInt8(Conversion.int8_to_uint8(value))

    def putUInt16(self, value):
        """ Append a uint16 to the buffer. """
        self.data.extend(Conversion.uint16_to_uint8_array(value))

    def putInt16(self, value):
        """ Append an int16 to the buffer. """
        self.putUInt16(Conversion.int16_to_uint16(value))

    def putUInt32(self, value):
        """ Append a uint32 to the buffer. """
        self.data.extend(Conversion.uint32_to_uint8_array(value))

    def putInt32(self, value):
        """ Append an int32 to the buffer. """
        self.putUInt32(Conversion.int32_to_uint32(value))

    def putFloat(self, value):
        """ Append a float to the buffer. """
        self.data.extend(Conversion.float_to_uint8_array(value))

    def putBytes(self, data):
        """ Append a list of bytes to the buffer. """
        self.data.extend(data)



import time

from crownstone_core.util.Conversion import Conversion


def obtainTimestamp(fullTimeStamp, lsb):
	timestampBytes = Conversion.uint32_to_uint8_array(int(fullTimeStamp))
	lsbBytes = Conversion.uint16_to_uint8_array(lsb)

	restoredTimestamp = Conversion.uint8_array_to_uint32([lsbBytes[0],lsbBytes[1],timestampBytes[2],timestampBytes[3]])

	return restoredTimestamp


def reconstructTimestamp(currentTimestamp, lsbTimestamp):
	# embed location data in the timestamp
	secondsFromGMT = round(time.time() - time.mktime(time.gmtime()))
	correctedTimestamp = currentTimestamp + secondsFromGMT

	# attempt restoration
	restoredTimestamp = obtainTimestamp(correctedTimestamp, lsbTimestamp)

	halfUInt16 = 0x7FFF # roughly 9 hours in seconds

	# correct for overflows, check for drift from current time
	delta = correctedTimestamp - restoredTimestamp

	if -halfUInt16 < delta < halfUInt16:
		return restoredTimestamp
	elif delta < -halfUInt16:
		restoredTimestamp = obtainTimestamp(correctedTimestamp - 0xFFFF, lsbTimestamp)
	elif delta > halfUInt16:
		restoredTimestamp = obtainTimestamp(correctedTimestamp + 0xFFFF, lsbTimestamp)

	return restoredTimestamp

def getCorrectedLocalTimestamp(currentTimestamp):
	secondsFromGMT = round(time.time() - time.mktime(time.gmtime()))
	correctedTimestamp = currentTimestamp + secondsFromGMT
	return correctedTimestamp


from crownstone_core.packets.CrownstoneErrors import CrownstoneErrors
from crownstone_core.packets.serviceDataParsers.parsers import parseOpCode5, parseOpCode6, parseOpCode4, parseOpCode3
from crownstone_core.protocol.BluenetTypes import DeviceType
from crownstone_core.protocol.SwitchState import SwitchState
from crownstone_core.util.EncryptionHandler import EncryptionHandler


class ServiceData:
    
    def __init__(self, data, unencrypted = False):
        self.opCode = 0
        self.dataType = 0
        self.crownstoneId = 0
        self.switchState = SwitchState(0)
        self.flagsBitmask = 0
        self.temperature = 0
        self.powerFactor = 1
        self.powerUsageReal = 0
        self.powerUsageApparent = 0
        self.accumulatedEnergy = 0
        self.setupMode = False
        self.stateOfExternalCrownstone = False
        self.data = None
        self.dataString = ""
        self.dimmerReady = False
        self.dimmingAllowed = False
        self.hasError = False
        self.switchLocked = False
        self.partialTimestamp = 0
        self.timestamp = -1
        self.validation = 0x00  # Will be 0xFAif it is set.
   
        self.errorTimestamp = 0
        self.errorsBitmask = 0
        self.errorMode = False
        self.timeIsSet = False
        self.switchCraftEnabled = False
        self.uniqueIdentifier = 0
   
        self.validData = False
        self.dataReadyForUse = False  # decryption is successful if this is true

        self.tapToToggleEnabled = False
        self.behaviourEnabled = True
        self.behaviourOverridden = False
        self.behaviourMasterHash = 0
    
        self.deviceType = DeviceType.UNDEFINED
        self.rssiOfExternalCrownstone = 0
 
        self.encryptedData = []
        self.encryptedDataStartIndex = 0
        
        self.data = data
        self.parse(unencrypted)

    def parse(self, unencrypted=False):
        self.validData = True
        if len(self.data) == 18:
            self.opCode = self.data[0]
            self.encryptedData = self.data[2:]
            self.encryptedDataStartIndex = 2
        elif len(self.data) == 17:
            self.opCode = self.data[0]
            self.encryptedData = self.data[1:]
            self.encryptedDataStartIndex = 1
        elif len(self.data) == 16 and unencrypted:
            self.encryptedData = self.data
            self.opCode = 7
        else:
            self.validData = False

        if self.validData:
            if self.opCode == 3:
                parseOpCode3(self, self.encryptedData)
            elif self.opCode == 4:
                parseOpCode4(self, self.encryptedData)
            elif self.opCode == 5 or self.opCode == 7:
                self.getDeviceTypeFromPublicData()
                parseOpCode5(self, self.encryptedData)
            elif self.opCode == 6:
                self.getDeviceTypeFromPublicData()
                parseOpCode6(self, self.encryptedData)
            else:
                self.getDeviceTypeFromPublicData()
                parseOpCode5(self, self.encryptedData)
            
    def getDeviceTypeFromPublicData(self):
        if len(self.data) == 18:
            if DeviceType.has_value(self.data[1]):
                self.deviceType = type

    def isInSetupMode(self):
        if not self.validData:
            return False
    
        return self.setupMode
    
    
    def getDictionary(self):
        errorsDictionary = CrownstoneErrors(self.errorsBitmask).getDictionary()
        
        returnDict = {}
        
        returnDict["opCode"]                    = self.opCode
        returnDict["dataType"]                  = self.dataType
        returnDict["stateOfExternalCrownstone"] = self.stateOfExternalCrownstone
        returnDict["hasError"]                  = self.hasError
        returnDict["setupMode"]                 = self.isInSetupMode()
        returnDict["id"]                        = self.crownstoneId
        returnDict["switchState"]               = self.switchState.raw
        returnDict["flagsBitmask"]              = self.flagsBitmask
        returnDict["temperature"]               = self.temperature
        returnDict["powerFactor"]               = self.powerFactor
        returnDict["powerUsageReal"]            = self.powerUsageReal
        returnDict["powerUsageApparent"]        = self.powerUsageApparent
        returnDict["accumulatedEnergy"]         = self.accumulatedEnergy
        returnDict["timestamp"]                 = self.timestamp
        returnDict["dimmerReady"]               = self.dimmerReady
        returnDict["dimmingAllowed"]            = self.dimmingAllowed
        returnDict["switchLocked"]              = self.switchLocked
        returnDict["switchCraftEnabled"]        = self.switchCraftEnabled
        returnDict["tapToToggleEnabled"]        = self.tapToToggleEnabled
        returnDict["behaviourEnabled"]          = self.behaviourEnabled
        returnDict["behaviourOverridden"]       = self.behaviourOverridden
        returnDict["behaviourMasterHash"]       = self.behaviourMasterHash

        returnDict["errorMode"]                 = self.errorMode
        returnDict["errors"]                    = errorsDictionary
    
        returnDict["uniqueElement"]             =  self.uniqueIdentifier
        returnDict["timeIsSet"]                 =  self.timeIsSet

        returnDict["rssiOfExternalCrownstone"]  = self.rssiOfExternalCrownstone
    
        return returnDict
    
    
    def getSummary(self, address):
        errorsDictionary = CrownstoneErrors(self.errorsBitmask).getDictionary()
    
        returnDict = {}
    
        returnDict["id"] = self.crownstoneId
        returnDict["address"] = address
        returnDict["setupMode"] = self.isInSetupMode()
        returnDict["switchState"] = self.switchState.raw
        returnDict["temperature"] = self.temperature
        returnDict["powerFactor"] = self.powerFactor
        returnDict["powerUsageReal"] = self.powerUsageReal
        returnDict["powerUsageApparent"] = self.powerUsageApparent
        returnDict["accumulatedEnergy"] = self.accumulatedEnergy
        returnDict["dimmerReady"] = self.dimmerReady
        returnDict["dimmingAllowed"] = self.dimmingAllowed
        returnDict["switchLocked"] = self.switchLocked
        returnDict["switchCraftEnabled"] = self.switchCraftEnabled
        returnDict["timeIsSet"] = self.timeIsSet
        returnDict["timestamp"] = self.timestamp
        returnDict["hasError"] = self.hasError
        returnDict["errorMode"] = self.errorMode
        returnDict["errors"] = errorsDictionary
    
        return returnDict


    def decrypt(self, keyHexString):
        if self.validData and len(self.encryptedData) == 16 and len(keyHexString) >= 16:
            if not self.setupMode:
                result = EncryptionHandler.decryptECB(self.encryptedData, keyHexString)

                for i in range(0, len(self.encryptedData)):
                    self.data[i + self.encryptedDataStartIndex] = result[i]

                self.parse()
            self.dataReadyForUse = True
        else:
            self.dataReadyForUse = False




#
#  BluenetLib
#
#  Created by Alex de Mulder on 22/10/2019.
#  Copyright © 2019 Alex de Mulder. All rights reserved.
#
from crownstone_core.packets.behaviour.BehaviourBase import BehaviourBase
from crownstone_core.packets.behaviour.BehaviourTypes import BehaviourType


class TwilightBehaviour(BehaviourBase):
    """
    Implements packet generation for TwilightBehaviour
    """
    def __init__(self, profileIndex=None, behaviourType=BehaviourType.twilight, intensity=None, activeDays=None, time=None, idOnCrownstone=None):
        super().__init__(profileIndex, behaviourType, intensity, activeDays, time, None, None, idOnCrownstone)





from crownstone_core.protocol.SwitchState import SwitchState
from crownstone_core.util.DataStepper import DataStepper
from crownstone_core.packets.debug.CommandSourcePacket import CommandSourcePacket

class SwitchHistoryListPacket:
	def __init__(self, data):
		self.list = []
		self.load(data)

	def load(self, data):
		"""
		Parses data buffer to set member variables.

		data : list of bytes

		Raises exception when parsing fails.
		"""
		streamBuf = DataStepper(data)
		count = streamBuf.getUInt8()
		self.list = []
		for i in range(0, count):
			self.list.append(SwitchHistoryItemPacket(streamBuf.getAmountOfBytes(SwitchHistoryItemPacket.size())))

	def toString(self):
		msg = "SwitchHistoryListPacket("
		msg += "list=["
		if (len(self.list)):
			msg += str(self.list[0])
		for i in range(1, len(self.list)):
			msg += ", " + str(self.list[i])
		msg += "]"
		msg += ")"
		return msg

	def __str__(self):
		return self.toString()

class SwitchHistoryItemPacket:
	def __init__(self, data):
		self.timestamp = 0     # Uint32
		self.switchCommand = 0 # Uint8
		self.switchState = SwitchState(0)
		self.source = None     # CommandSourcePacket
		self.load(data)

	def load(self, data):
		"""
		Parses data buffer to set member variables.

		data : list of bytes

		Raises exception when parsing fails.
		"""
		streamBuf = DataStepper(data)
		self.timestamp = streamBuf.getUInt32()
		self.switchCommand = streamBuf.getUInt8()
		self.switchState = SwitchState(streamBuf.getUInt8())
		self.source = CommandSourcePacket(streamBuf.getAmountOfBytes(CommandSourcePacket.size()))

	@staticmethod
	def size():
		return 4 + 1 + 1 + CommandSourcePacket.size()

	def toString(self):
		msg = "CommandSourcePacket("
		msg += "timestamp=" + str(self.timestamp)
		msg += " switchCommand=" + str(self.switchCommand)
		msg += " switchState=" + str(self.switchState)
		msg += " source=" + str(self.source)
		msg += ")"
		return msg

	def __str__(self):
		return self.toString()



import time


prev  = time.time()
start = time.time()

def gt():
    global prev
    t = time.time()
    strD = "{:.3f} {:.3f} ".format(t,t - prev)
    prev = t
    return strD

def dt():
    global prev
    t = time.time()
    strD = "{:.3f} ".format(t - prev)
    prev = t
    return strD

def tfs():
    global start
    t = time.time()
    strD = "{:.3f} ".format(t - start)
    return strD


from crownstone_core.packets.serviceDataParsers.opCode3.opCode3_type0 import parseOpCode3_type0


def parseOpCode3_type2(serviceData, data):
    if len(data) == 16:
        parseOpCode3_type0(serviceData, data)
    
        # apply differences between type 0 and type 2
        serviceData.stateOfExternalCrownstone = True




"""
 We will assume that there is a dictionary format which is used to provide the system with the behaviour data. We will raise an error on incorrect format. Otherwise we will parse it and return the Behaviour object.
 Expected format:
 {
    type: "BEHAVIOUR" | "TWILIGHT",
    data: {
        action: { data: number },
        time: TIME_PARSER_INPUT,
        presence: PRESENCE_PARSER_INPUT
        endCondition: ENDCONDITION_PARSER_INPUT
    },
    activeDays: {
      Mon: boolean,
      Tue: boolean,
      Wed: boolean,
      Thu: boolean,
      Fri: boolean,
      Sat: boolean,
      Sun: boolean
    },
    idOnCrownstone: number,
    profileIndex: number
 }
 """
import json

from crownstone_core.Exceptions import CrownstoneError
from crownstone_core.packets.behaviour.ActiveDays import ActiveDays
from crownstone_core.packets.behaviour.BehaviourBase import BehaviourBase
from crownstone_core.packets.behaviour.BehaviourTypes import BehaviourType, DAY_START_TIME_SECONDS_SINCE_MIDNIGHT
from crownstone_core.packets.behaviour.PresenceDescription import BehaviourPresenceType, BehaviourPresence
from crownstone_core.packets.behaviour.TimeDescription import BehaviourTimeContainer, BehaviourTime, BehaviourTimeType


def getFromDict(dict, key):
    val = None
    if key in dict: val = dict[key]
    return val

def BehaviourDictionaryParser(dict, dayStartTimeSecondsSinceMidnight = DAY_START_TIME_SECONDS_SINCE_MIDNIGHT):
    # optional variables
    profileIndex   = getFromDict(dict, "profileIndex")                                  
    typeString     = getFromDict(dict, "type")
    data           = getFromDict(dict, "data")                          
    activeDays     = getFromDict(dict, "activeDays")                                  
    idOnCrownstone = getFromDict(dict, "idOnCrownstone")                                      
    
    if profileIndex  is None: raise CrownstoneError.PROFILE_INDEX_MISSING
    if typeString    is None: raise CrownstoneError.TYPE_MISSING
    if data          is None: raise CrownstoneError.DATA_MISSING
    if activeDays    is None: raise CrownstoneError.ACTIVE_DAYS_MISSING
    
    behaviourType = BehaviourType.behaviour
    if   typeString == "BEHAVIOUR" : behaviourType = BehaviourType.behaviour
    elif typeString == "TWILIGHT"  : behaviourType = BehaviourType.twilight

    if type(data) == str:
        data = json.loads(data)

    # optional variables
    actionDict   = getFromDict(data, "action")
    timeDict     = getFromDict(data, "time")
    presence     = getFromDict(data, "presence")
    endCondition = getFromDict(data, "endCondition")

    if actionDict is None: raise CrownstoneError.BEHAVIOUR_ACTION_MISSING
    if timeDict   is None: raise CrownstoneError.BEHAVIOUR_TIME_MISSING

  
    intensity = getFromDict(actionDict, "data")
    
    if intensity is None: raise CrownstoneError.BEHAVIOUR_INTENSITY_MISSING 
    
    activeDayObject = ActiveDayParser(activeDays)
    timeObject      = TimeParser(timeDict, dayStartTimeSecondsSinceMidnight)
    
    behaviour = BehaviourBase(
        profileIndex  = profileIndex,
        behaviourType = behaviourType,
        intensity     = intensity,
        activeDays    = activeDayObject,
        time          = timeObject
    )
    
    if idOnCrownstone is not None:
        behaviour.idOnCrownstone = idOnCrownstone
        
    if presence is not None:
        if behaviourType == BehaviourType.twilight :
            raise CrownstoneError.TWILIGHT_CANT_HAVE_PRESENCE
        
        presenceObject = PresenceParser(presence)
        behaviour.presence = presenceObject
    
    
    if endCondition is not None:
        if behaviourType == BehaviourType.twilight :
            raise CrownstoneError.TWILIGHT_CANT_HAVE_END_CONDITION
        
        endConditionObject = EndConditionParser(endCondition)
        behaviour.endCondition = endConditionObject
        behaviour.type = BehaviourType.smartTimer
    
    
    return behaviour


"""
 There are a few possible formats. We will parse and validate
"""
def ActiveDayParser(dict):
    Monday    = getFromDict(dict, "Mon")
    Tuesday   = getFromDict(dict, "Tue")
    Wednesday = getFromDict(dict, "Wed")
    Thursday  = getFromDict(dict, "Thu")
    Friday    = getFromDict(dict, "Fri")
    Saturday  = getFromDict(dict, "Sat")
    Sunday    = getFromDict(dict, "Sun")
    
    activeDays = ActiveDays()
    
    if Monday is not None and Tuesday is not None and Wednesday is not None and Thursday is not None and Friday is not None and Saturday is not None and Sunday is not None:
        activeDays.Monday    = Monday
        activeDays.Tuesday   = Tuesday
        activeDays.Wednesday = Wednesday
        activeDays.Thursday  = Thursday
        activeDays.Friday    = Friday
        activeDays.Saturday  = Saturday
        activeDays.Sunday    = Sunday
        
        if activeDays.getMask() == 0:
            raise CrownstoneError.NO_ACTIVE_DAYS
        
        return activeDays
    else :
        raise CrownstoneError.ACTIVE_DAYS_INVALID
    


"""
 There are a few possible formats. We will parse and validate
"""
def TimeParser(dict, dayStartTimeSecondsSinceMidnight):
    type = getFromDict(dict, "type")

    if type is None: raise CrownstoneError.NO_TIME_TYPE

    if type == "ALL_DAY":
        return BehaviourTimeContainer(
            BehaviourTime().fromType(BehaviourTimeType.afterMidnight, dayStartTimeSecondsSinceMidnight),
            BehaviourTime().fromType(BehaviourTimeType.afterMidnight, dayStartTimeSecondsSinceMidnight)
        )
    
    elif type == "RANGE":
        fromTime = getFromDict(dict, "from")
        toTime   = getFromDict(dict, "to")

        if fromTime is None: raise CrownstoneError.MISSING_FROM_TIME
        if toTime   is None: raise CrownstoneError.MISSING_TO_TIME

        fromTimeType = getFromDict(fromTime, "type")
        toTimeType   = getFromDict(toTime,   "type")

        if fromTimeType is None: raise CrownstoneError.MISSING_FROM_TIME_TYPE
        if toTimeType   is None: raise CrownstoneError.MISSING_TO_TIME_TYPE

        fromResult = None
        toResult   = None
        
        if fromTimeType == "CLOCK":
            fromData = getFromDict(fromTime, "data")
        
            if fromData is None: raise CrownstoneError.MISSING_FROM_TIME_DATA

            hours   = getFromDict(fromData, "hours")
            minutes = getFromDict(fromData, "minutes")

            if hours   is None: raise CrownstoneError.MISSING_FROM_TIME_DATA
            if minutes is None: raise CrownstoneError.MISSING_FROM_TIME_DATA

            fromResult = BehaviourTime().fromTime(hours, minutes)

        elif fromTimeType == "SUNSET":
            offsetSeconds = 0
            offsetMinutes = getFromDict(fromTime, "offsetMinutes")
            if offsetMinutes is not None:
                offsetSeconds = 60*offsetMinutes

            fromResult = BehaviourTime().fromType(BehaviourTimeType.afterSunset, offsetSeconds)
        
        elif fromTimeType == "SUNRISE":
            offsetSeconds = 0
            offsetMinutes = getFromDict(fromTime, "offsetMinutes")
            if offsetMinutes is not None:
                offsetSeconds = 60 * offsetMinutes
            
            fromResult = BehaviourTime().fromType(BehaviourTimeType.afterSunrise, offsetSeconds)
        
        else:
            raise CrownstoneError.INVALID_TIME_FROM_TYPE
        
        
        
        if toTimeType == "CLOCK":
            toData = getFromDict(toTime, "data")

            if toData is None: raise CrownstoneError.MISSING_TO_TIME_DATA

            hours = getFromDict(toData, "hours")
            minutes = getFromDict(toData, "minutes")

            if hours is None: raise CrownstoneError.INVALID_TO_DATA
            if minutes is None: raise CrownstoneError.INVALID_TO_DATA

            toResult = BehaviourTime().fromTime(hours, minutes)

        elif toTimeType == "SUNSET":
            offsetSeconds = 0
            offsetMinutes = getFromDict(toTime, "offsetMinutes")
            if offsetMinutes is not None:
                offsetSeconds = 60 * offsetMinutes

            toResult = BehaviourTime().fromType(BehaviourTimeType.afterSunset, offsetSeconds)

        elif toTimeType == "SUNRISE":
            offsetSeconds = 0
            offsetMinutes = getFromDict(toTime, "offsetMinutes")
            if offsetMinutes is not None:
                offsetSeconds = 60 * offsetMinutes

            toResult = BehaviourTime().fromType(BehaviourTimeType.afterSunrise, offsetSeconds)
        
        else:
            raise CrownstoneError.INVALID_TIME_TO_TYPE
        
        return BehaviourTimeContainer(fromResult, toResult)
    else:
        raise CrownstoneError.INVALID_TIME_TYPE
    
    


"""
 There are a few possible formats. We will parse and validate
"""
def PresenceParser(dict):
    type = getFromDict(dict, "type")
    if type is None: raise CrownstoneError.NO_PRESENCE_TYPE

    if type == "IGNORE":
        return BehaviourPresence()
    elif type == "SOMEBODY" or type == "NOBODY":
        data = getFromDict(dict, "data")
        if data is None: raise CrownstoneError.NO_PRESENCE_DATA

        delay = getFromDict(dict, "delay")
        if delay is None: raise CrownstoneError.NO_PRESENCE_DELAY

        dataType = getFromDict(data, "type")
        if dataType is None: raise CrownstoneError.NO_PRESENCE_TYPE

        if dataType == "SPHERE":
            if type == "SOMEBODY":
                return BehaviourPresence().setSpherePresence(BehaviourPresenceType.someoneInSphere, delayInSeconds = delay)
            else :
                return BehaviourPresence().setSpherePresence(BehaviourPresenceType.nobodyInSphere, delayInSeconds = delay)

        elif dataType == "LOCATION":
            locationIdArray = getFromDict(data, "locationIds")
            if locationIdArray is None: raise CrownstoneError.NO_PRESENCE_LOCATION_IDS

            if type == "SOMEBODY":
                return BehaviourPresence().setLocationPresence(BehaviourPresenceType.somoneInLocation, locationIds=locationIdArray, delayInSeconds=delay)
            
            else :
                return BehaviourPresence().setLocationPresence(BehaviourPresenceType.nobodyInLocation, locationIds=locationIdArray, delayInSeconds=delay)
        else :
            raise CrownstoneError.NO_PRESENCE_DATA
    else :
        raise CrownstoneError.INVALID_PRESENCE_TYPE
    
    
    


"""
 There are a few possible formats. We will parse and validate
"""
def EndConditionParser(dict):
    type     = getFromDict(dict, "type")
    presence = getFromDict(dict, "presence")
    if type     is None: raise CrownstoneError.NO_END_CONDITION_TYPE
    if presence is None: raise CrownstoneError.NO_END_CONDITION_PRESENCE

    presenceObject = PresenceParser(presence)
    
    return presenceObject




from crownstone_core.Exceptions import CrownstoneError
from crownstone_core.util.Conversion import Conversion

# TODO: rename to DataParser or so?
class DataStepper :
    """
    Class to parse a data buffer.

    It keeps up a position in the buffer, so you can do consecutive calls to get.

    Function calls will raise an exception on error.
    """
    def __init__(self, data):
        """
        Constructor

        data : list of bytes, being the buffer.
        """
        self.data = data
        self.length = len(self.data)

        self.position = 0
        self.markPosition = 0


    def getUInt8(self):
        """ Get a uint8 from the current position, and advance the position. """
        return self._request(1)[0]

    def getInt8(self):
        """ Get an int8 from the current position, and advance the position. """
        return Conversion.uint8_to_int8(self.getUInt8())

    def getUInt16(self):
        """ Get a uint16 from the current position, and advance the position. """
        return Conversion.uint8_array_to_uint16(self._request(2))

    def getInt16(self):
        """ Get an int16 from the current position, and advance the position. """
        return Conversion.uint8_array_to_int16(self._request(2))

    def getUInt32(self):
        """ Get a uint32 from the current position, and advance the position. """
        return Conversion.uint8_array_to_uint32(self._request(4))

    def getInt32(self):
        """ Get an int32 from the current position, and advance the position. """
        return Conversion.uint8_array_to_int32(self._request(4))

    def getUInt64(self):
        """ Get a uint64 from the current position, and advance the position. """
        return Conversion.uint8_array_to_uint64(self._request(8))

    def getFloat(self):
        """ Get a float from the current position, and advance the position. """
        return Conversion.uint8_array_to_float(self._request(4))

    def skip(self, count=1):
        """ Advance the position by N bytes. """
        self._request(count)

    # TODO: rename to getRawBytes() or so, now it looks as if you're getting the buffer size.
    def getAmountOfBytes(self, amount):
        """ Get N bytes of data. """
        return self._request(amount)

    def getRemainingBytes(self):
        """ Get remaining bytes of data. """
        return self._request(self.remaining())

    def mark(self):
        """ Mark the current position, so you can return to it later with reset(). """
        self.markPosition = self.position

    def reset(self):
        """ Reset the position to the last marked position. """
        self.position = self.markPosition

    def remaining(self):
        """ Return the number of bytes remaining in the buffer. """
        return self.length - self.position

    def _request(self, size):
        if self.position + size <= self.length:
            start = self.position
            self.position += size
            return self.data[start:self.position]
        else:
            raise CrownstoneError.INVALID_DATA_LENGTH







from crownstone_core.packets.behaviour.BehaviourBase import BehaviourBase
from crownstone_core.packets.behaviour.BehaviourTypes import BehaviourType
from crownstone_core.packets.behaviour.PresenceDescription import BehaviourPresence, BehaviourPresenceType, \
    DEFAULT_PRESENCE_DELAY

DEFAULT_PRESENCE = BehaviourPresence().setSpherePresence(BehaviourPresenceType.someoneInSphere)


class SwitchBehaviour(BehaviourBase):
    """
    Implements packet generation for SwitchBehaviours
    """

    def __init__(self, profileIndex=0, behaviourType=BehaviourType.behaviour, intensity=None, activeDays=None,
                 time=None, presence=None, endCondition=None, idOnCrownstone=None):
        super().__init__(profileIndex, behaviourType, intensity, activeDays, time, presence, endCondition,
                         idOnCrownstone)

    def ignorePresence(self):
        self.presence = None
        return self

    def setPresenceIgnore(self):
        return self.ignorePresence()

    def setPresenceSomebody(self):
        self.setPresenceSomebodyInSphere()
        return self

    def setPresenceNobody(self):
        self.setPresenceNobodyInSphere()
        return self

    def setPresenceSomebodyInSphere(self):
        self.presence = BehaviourPresence().setSpherePresence(BehaviourPresenceType.someoneInSphere)
        return self

    def setPresenceNobodyInSphere(self):
        self.presence = BehaviourPresence().setSpherePresence(BehaviourPresenceType.nobodyInSphere)
        return self

    def setPresenceInSphere(self):
        self.setPresenceSomebodyInSphere()
        return self

    def setPresenceInLocations(self, locationIds):
        self.setPresenceSomebodyInLocations(locationIds)
        return self

    def setPresenceSomebodyInLocations(self, locationIds, delay=DEFAULT_PRESENCE_DELAY):
        self.presence = BehaviourPresence().setLocationPresence(BehaviourPresenceType.somoneInLocation, locationIds,
                                                                delay)
        return self

    def setPresenceNobodyInLocations(self, locationIds, delay=DEFAULT_PRESENCE_DELAY):
        self.presence = BehaviourPresence().setLocationPresence(BehaviourPresenceType.nobodyInLocation, locationIds,
                                                                delay)
        return self

    def getPacket(self):
        arr = super().getPacket()

        if self.presence is not None:
            arr += self.presence.getPacket()
        else:
            anypresence = BehaviourPresence()
            arr += anypresence.getPacket()

        return arr




import json
import sys
import os

__location__ = sys.path[0]

class JsonFileStore:
    
    
    
    def __init__(self, filename):
        self.memoryStore = {}
        self.filename = filename
        self._loadData()
    
    
    def _loadData(self):
        fileHandle = open(os.path.join(__location__, self.filename),'r')
        data = fileHandle.read()
        self.memoryStore = json.loads(data)
        fileHandle.close()
        
    
    def addEntry(self, key, value):
        self.memoryStore[key] = value
        self._persist()


    def updateEntry(self, key, value):
        self.addEntry(key, value)


    def removeEntry(self, key):
        del self.memoryStore[key]
     
     
    def getData(self):
        return self.memoryStore
    
     
    def _persist(self):
        jsonString = json.dumps(self.memoryStore)
        
        fileHandle = open(self.filename, 'w')
        fileHandle.write(jsonString)
        fileHandle.close()
        
    
    def clear(self):
        fileHandle = open(self.filename, 'w')
        fileHandle.write("")
        fileHandle.close()
        self.memoryStore = {}
    




from enum import IntEnum
from typing import List


# broadcast to all:
# value: 1
#
# broadcast to all, but retry until ID's in list have acked or timeout
# value: 3
#
# 1:1 message to N crownstones with acks (only N = 1 supported for now)
# value: 2

class MeshModes(IntEnum):
    BROADCAST           = 1
    SINGLE_TARGET_ACKED = 2
    BROADCAST_ACKED_IDS = 3


class _MeshCommandPacket:

    def __init__(self, crownstoneIds : List[int], payload, mesh_command_mode: MeshModes, timeout_or_transmissions):
        self.type = 0 # reserved
        self.crownstoneIds = crownstoneIds
        self.payload = payload
        self.flags = mesh_command_mode.value
        self.timeout_or_transmissions = timeout_or_transmissions

    def getPacket(self):
        packet = []
        packet.append(self.type)
        packet.append(self.flags)
        packet.append(self.timeout_or_transmissions)
        packet.append(len(self.crownstoneIds))
        packet += self.crownstoneIds
        packet += self.payload

        return packet

class MeshBroadcastPacket(_MeshCommandPacket):

    def __init__(self, payload, number_of_transmissions : int = 0):
        """
            number_of_transmissions 0 uses the default number_of_transmissions which is 3
        """
        super().__init__([], payload, MeshModes.BROADCAST, number_of_transmissions)

class MeshSetStatePacket(_MeshCommandPacket):

    def __init__(self, crownstoneId : int, setStatePacket, timeout_seconds : int = 0 ):
        """
        timeout_seconds 0 uses the default timeout of 10 seconds
        """
        super().__init__([crownstoneId], setStatePacket, MeshModes.SINGLE_TARGET_ACKED, timeout_seconds)

class MeshBroadcastAckedPacket(_MeshCommandPacket):
    """
    This is currently only supported for type setIBeaconConfig
    """
    def __init__(self, crownstoneIds: List[int], payload, timeout_seconds : int = 0):
        """
        timeout_seconds 0 uses the default timeout of 10 seconds
        """
        super().__init__(crownstoneIds, payload, MeshModes.BROADCAST_ACKED_IDS, timeout_seconds)


class StoneMultiSwitchPacket:

    def __init__(self, crownstoneId: int, switchVal: int):
        """
        :param crownstoneId:
        :param switchVal:  percentage [0..100] or special value (SwitchValSpecial).

        """
        self.crownstoneId = crownstoneId
        self.state = switchVal


    def getPacket(self):
        packet = []
        packet.append(self.crownstoneId)
        packet.append(self.state)

        return packet


class MeshMultiSwitchPacket:

    def __init__(self, packets=None):
        if packets is None:
            packets = []
        self.packets = packets

    def getPacket(self):
        packet = []
        packet.append(len(self.packets))
        for stonePacket in self.packets:
            packet += stonePacket.getPacket()

        return packet


from crownstone_core.protocol.BlePackets import ControlPacket, FactoryResetPacket
from crownstone_core.protocol.BluenetTypes import ControlType
from crownstone_core.util.Conversion import Conversion
from crownstone_core.util.BufferFiller import BufferFiller

class ControlPacketsGenerator:

	@staticmethod
	def getFactoryResetPacket():
		return Conversion.uint32_to_uint8_array(0xdeadbeef)


	@staticmethod
	def getCommandFactoryResetPacket():
		return FactoryResetPacket().getPacket()

	@staticmethod
	def getSwitchCommandPacket(switchVal: int):
		"""
		:param switchVal: percentage [0..100] or special value (SwitchValSpecial).
		"""
		return ControlPacket(ControlType.SWITCH).loadUInt8(switchVal).getPacket()

	@staticmethod
	def getResetPacket():
		return ControlPacket(ControlType.RESET).getPacket()

	@staticmethod
	def getPutInDFUPacket():
		return ControlPacket(ControlType.GOTO_DFU).getPacket()

	@staticmethod
	def getDisconnectPacket():
		return ControlPacket(ControlType.DISCONNECT).getPacket()

	@staticmethod
	def getRelaySwitchPacket(turnOn: bool):
		"""
		:param turnOn: True to turn relay on.
		"""
		switchVal = 0
		if turnOn:
			switchVal = 1

		return ControlPacket(ControlType.RELAY).loadUInt8(switchVal).getPacket()

	@staticmethod
	def getDimmerSwitchPacket(intensity: int):
		"""
		:param intensity: percentage [0..100]
		"""
		return ControlPacket(ControlType.PWM).loadUInt8(intensity).getPacket()


	@staticmethod
	def getResetErrorPacket(errorMask):
		return ControlPacket(ControlType.RESET_ERRORS).loadUInt32(errorMask).getPacket()

	@staticmethod
	def getSetTimePacket(time):
		"""
		This is a LOCAL timestamp since epoch in seconds

		so if you live in GMT + 1 add 3600 to the timestamp
		:param time:
		:return:
		"""
		return ControlPacket(ControlType.SET_TIME).loadUInt32(time).getPacket()

	@staticmethod
	def getAllowDimmingPacket(allow: bool):
		"""

		:param allow: bool
		:return:
		"""

		allowByte = 0
		if allow:
			allowByte = 1

		return ControlPacket(ControlType.ALLOW_DIMMING).loadUInt8(allowByte).getPacket()

	@staticmethod
	def getLockSwitchPacket(lock: bool):
		"""
		:param lock: bool
		:return:
		"""

		lockByte = 0
		if lock:
			lockByte = 1

		return ControlPacket(ControlType.LOCK_SWITCH).loadUInt8(lockByte).getPacket()

	@staticmethod
	def getMicroAppUploadPacket(microapp_packet):
		"""
        :param protocol:        uint8 number
        :param app_id:          uint8 number
        :param opcode:          uint8 number
        :param index:           uint8 number
        :param checksum:        uint16 number
        :param buffer:          byteString (no conversion required)
        :return:
        """
		data = []
		data.append(microapp_packet.protocol)
		data.append(microapp_packet.app_id)
		data.append(microapp_packet.opcode)
		data.append(microapp_packet.index)
		data += Conversion.uint16_to_uint8_array(microapp_packet.checksum)
		data += list(microapp_packet.buffer)
		return ControlPacket(ControlType.MICROAPP).loadByteArray(data).getPacket()

	@staticmethod
	def getMicroAppRequestPacket(packet):
		"""
        :param protocol:        uint8 number
        :param app_id:          uint8 number
        :param opcode:          uint8 number
        :param size:            uint16 number
        :param count:           uint8 number
        :param chunk_size:      uint8 number
        :return:
        """
		data = []
		data.append(packet.protocol)
		data.append(packet.app_id)
		data.append(packet.opcode)
		data += Conversion.uint16_to_uint8_array(packet.size)
		data.append(packet.count)
		data.append(packet.chunk_size)
		return ControlPacket(ControlType.MICROAPP).loadByteArray(data).getPacket()

	@staticmethod
	def getMicroAppValidatePacket(packet):
		"""
        :param protocol:        uint8 number
        :param app_id:          uint8 number
        :param opcode:          uint8 number
        :param size:            uint16 number
        :param checksum:        uint16 number
        :return:
        """
		data = []
		data.append(packet.protocol)
		data.append(packet.app_id)
		data.append(packet.opcode)
		data += Conversion.uint16_to_uint8_array(packet.size)
		data += Conversion.uint16_to_uint8_array(packet.checksum)
		return ControlPacket(ControlType.MICROAPP).loadByteArray(data).getPacket()



	@staticmethod
	def getMicroAppEnablePacket(microapp_packet):
		"""
        :param protocol:        uint8 number
        :param app_id:          uint8 number
        :param opcode:          uint8 number
        :param param0:          uint16 number
        :return:
        """
		data = []
		data.append(microapp_packet.protocol)
		data.append(microapp_packet.app_id)
		data.append(microapp_packet.opcode)
		data += Conversion.uint16_to_uint8_array(microapp_packet.param0)
		return ControlPacket(ControlType.MICROAPP).loadByteArray(data).getPacket()

	@staticmethod
	def getSetupPacket(
		crownstoneId: int,
		sphereId: int,
		adminKey,
		memberKey,
		basicKey,
		serviceDataKey,
		localizationKey,
		meshDeviceKey,
		meshAppKey,
		meshNetworkKey,
		ibeaconUUID: str,
		ibeaconMajor: int,
		ibeaconMinor: int
	):
		"""
		:param crownstoneId:  		uint8 number
		:param sphereId:  	     	uint8 number
		:param adminKey:      		byteString (no conversion required)
		:param memberKey:     		byteString (no conversion required)
		:param basicKey:      		byteString (no conversion required)
		:param serviceDataKey: 	    byteString (no conversion required)
		:param localizationKey: 	byteString (no conversion required)
		:param meshDeviceKey: 	    byteString (no conversion required)
		:param meshAppKey: 	        byteString (no conversion required)
		:param meshNetworkKey: 	    byteString (no conversion required)
		:param ibeaconUUID: 		string  (ie. "1843423e-e175-4af0-a2e4-31e32f729a8a")
		:param ibeaconMajor:        uint16 number
		:param ibeaconMinor:        uint16 number
		:return:
		"""
		data = []
		data.append(crownstoneId)
		data.append(sphereId)

		data += list(adminKey)
		data += list(memberKey)
		data += list(basicKey)
		data += list(serviceDataKey)
		data += list(localizationKey)

		MDKey = meshDeviceKey
		if type(meshDeviceKey) is str:
			MDKey = Conversion.ascii_or_hex_string_to_16_byte_array(meshDeviceKey)

		data += list(MDKey)
		data += list(meshAppKey)
		data += list(meshNetworkKey)

		data += Conversion.ibeaconUUIDString_to_reversed_uint8_array(ibeaconUUID)
		data += Conversion.uint16_to_uint8_array(ibeaconMajor)
		data += Conversion.uint16_to_uint8_array(ibeaconMinor)

		return ControlPacket(ControlType.SETUP).loadByteArray(data).getPacket()


	@staticmethod
	def getIBeaconConfigIdPacket(id, timestamp, interval):
		data = []
		data.append(id)
		data += Conversion.uint32_to_uint8_array(timestamp)
		data += Conversion.uint16_to_uint8_array(interval)

		return ControlPacket(ControlType.SET_IBEACON_CONFIG_ID).loadByteArray(data).getPacket()

	@staticmethod
	def getPowerSamplesRequestPacket(samplesType, index):
		buffer = BufferFiller()
		buffer.putUInt8(samplesType)
		buffer.putUInt8(index)
		data = buffer.getBuffer()
		return ControlPacket(ControlType.GET_POWER_SAMPLES).loadByteArray(data).getPacket()



from crownstone_core.protocol.BluenetTypes import ControlType, SetPersistenceMode, GetPersistenceMode
from crownstone_core.util.Conversion import Conversion

SUPPORTED_PROTOCOL_VERSION = 5

# TODO: this isn't just for BLE as the name suggests.
class BLEPacket:

    def __init__(self, packetType):
        self.lengthAsUint8Array = [0, 0]
        self.payload = []
        self.type = packetType.value

    def loadKey(self, keyString):
        self.payload = Conversion.ascii_or_hex_string_to_16_byte_array(keyString)
        return self._process()
    
    def loadString(self, string):
        self.payload = Conversion.string_to_uint8_array(string)
        return self._process()

    def loadInteger(self, integer):
        self.payload = [Conversion.int8_to_uint8(integer)]
        return self._process()

    def loadUInt8(self, uint8):
        self.payload = [uint8]
        return self._process()

    def loadUInt16(self, uint16):
        self.payload = Conversion.uint16_to_uint8_array(uint16)
        return self._process()

    def loadUInt32(self, uint32):
        self.payload = Conversion.uint32_to_uint8_array(uint32)
        return self._process()

    def loadFloat(self, float):
        self.payload = Conversion.float_to_uint8_array(float)
        return self._process()

    def loadByteArray(self, byteArray):
        self.payload = byteArray
        return self._process()

    def appendByteArray(self, byteArray):
        self.payload += byteArray
        return self._process()

    def _process(self):
        self.length = len(self.payload)
        self.lengthAsUint8Array = Conversion.uint16_to_uint8_array(len(self.payload))
        return self

    def getPacket(self):
        packet = [SUPPORTED_PROTOCOL_VERSION]
        packet += Conversion.uint16_to_uint8_array(self.type)
        packet += self.lengthAsUint8Array
        packet += self.payload

        return packet

class ControlPacket(BLEPacket):

    def __init__(self, packetType):
        super().__init__(packetType)


class FactoryResetPacket(ControlPacket):

    def __init__(self):
        super().__init__(ControlType.FACTORY_RESET)
        self.loadUInt32(0xdeadbeef)

class ControlStateGetPacket(ControlPacket):

    def __init__(self, stateType, id = 0, persistenceMode = GetPersistenceMode.CURRENT):
        super().__init__(ControlType.GET_STATE)
        self.id = id
        self.loadUInt16(stateType)
        self.persistenceMode = persistenceMode.value


    def getPacket(self):
        # TODO: make use of the base class, now there is a double implementation of the control packet.
        arr = [SUPPORTED_PROTOCOL_VERSION]
        arr += Conversion.uint16_to_uint8_array(self.type)
        arr += Conversion.uint16_to_uint8_array(self.length + 6) # the + 2 is for the stateType uint16, +2 for the id, +2 for persistenceMode and reserved
        arr += self.payload # this is the state type
        arr += Conversion.uint16_to_uint8_array(self.id)
        arr.append(self.persistenceMode)
        arr.append(0)
        return arr


class ControlStateSetPacket(ControlPacket):

    def __init__(self, stateType, id=0, persistenceMode = SetPersistenceMode.STORED):
        super().__init__(ControlType.SET_STATE)
        self.stateType = stateType
        self.id = id
        self.persistenceMode = persistenceMode.value


    def getPacket(self):
        # TODO: make use of the base class, now there is a double implementation of the control packet.
        arr = [SUPPORTED_PROTOCOL_VERSION]
        arr += Conversion.uint16_to_uint8_array(self.type)
        arr += Conversion.uint16_to_uint8_array(self.length + 6) # the + 2 is for the stateType uint16, +2 for the id, +2 for persistenceMode and reserved
        arr += Conversion.uint16_to_uint8_array(self.stateType)
        arr += Conversion.uint16_to_uint8_array(self.id)
        arr.append(self.persistenceMode)
        arr.append(0)
        arr += self.payload
        return arr

